import os
from typing import Optional, Union, List, Tuple

import numpy as np

from astropy.nddata import VarianceUncertainty  # type: ignore
from astropy.io import fits  # type: ignore
from astropy.table import Table  # type: ignore
from astropy import wcs  # type: ignore
from astropy import units  # type: ignore
from specutils import Spectrum1D  # type: ignore
from specutils.io.registers import identify_spectrum_format  # type: ignore
from specutils.io.registers import data_loader  # type: ignore

KNOWN_SPEC_EXT_NAMES = ['spec', 'spectrum', 'flux', 'data', 'sci', 'science']
KNOWN_VARIANCE_EXT_NAMES = ['stat', 'stats', 'var', 'variance', 'noise', 'err']
KNOWN_INVAR_EXT_NAMES = ['ivar', 'ivariance']
KNOWN_MASK_EXT_NAMES = ['mask', 'platemask', 'footprint', 'dq', 'nan_mask']
KNOWN_WAVE_EXT_NAMES = ['wave', 'wavelenght', 'lambda', 'lam']
KNOWN_RCURVE_EXT_NAMES = ['r', 'reso', 'resolution', 'rcurve', 'wd']
KNOWN_RGB_EXT_NAMES = ['r', 'g', 'b', 'red', 'green', 'blue']


def identifySpecexFits(origin, *args, **kwargs) -> bool:
    """Identify spectra extracted with python-specex."""
    if not (
        isinstance(args[0], str) and
        os.path.splitext(args[0].lower())[1] == '.fits'
    ):
        return False

    valid_id_keys = [
        f"{i}{j}"
        for i in ['', 'OBJ', 'OBJ_', 'TARGET', 'TARGET_', 'OBJECT', 'OBJECT_']
        for j in ['ID', 'NUMBER', 'UID', 'UUID', '']
    ]

    with fits.open(args[0]) as hdul:
        for hdu in hdul:
            for key in valid_id_keys:
                try:
                    _ = hdu.header[key]
                except KeyError:
                    continue
                else:
                    return True
    return False


@data_loader(
    label="specex-1d",
    identifier=identifySpecexFits,
    extensions=['fits']
)
def specexFitsLoader(
    file_name: str,
    flux_hdu_index: Optional[Union[str, int]] = None,
    resolution: Union[int, float] = 1,
    **kwargs
) -> Spectrum1D:
    """
    Load a python-specex spectrum.

    :param file_name:
        The path of the FITS file to read.
    :param flux_hdu_index:
        The extension name or index from which to read the spectrum
    :param resolution:
        The spectral resolution parameter
    :return: A Spectrum1D object.
    """
    def getHDU(
        hdul: fits.HDUList,
        valid_names: List[str],
        index: Optional[Union[str, int]] = None
    ):
        if index is None:
            for hdu in hdul:
                if hdu.name.lower() in valid_names:
                    return hdu
            else:
                return None
        else:
            return hdul[index]

    with fits.open(file_name, **kwargs) as hdulist:
        flux_hdu = getHDU(hdulist, KNOWN_SPEC_EXT_NAMES, index=flux_hdu_index)
        var_hdu = getHDU(hdulist, KNOWN_VARIANCE_EXT_NAMES)
        mask_hdu = getHDU(hdulist, KNOWN_MASK_EXT_NAMES)
        rc_hdu = getHDU(hdulist, KNOWN_RCURVE_EXT_NAMES)

        valid_id_keys: List[str] = [
            f"{i}{j}"
            for i in ['', 'OBJ', 'OBJ_', 'TARGET', 'TARGET_']
            for j in ['ID', 'NUMBER', 'UID', 'UUID']
        ]

        obj_id = None
        for key in valid_id_keys:
            try:
                obj_id = hdulist[0].header[key]
            except KeyError:
                continue
            else:
                break

        if flux_hdu is None:
            raise ValueError("Cannot find flux data")
        flux = flux_hdu.data

        if var_hdu is None:
            var = np.ones_like(flux)
        else:
            var = var_hdu.data

        if mask_hdu is None:
            mask = None
        else:
            mask = mask_hdu.data

        if rc_hdu is None:
            rc = None
        else:
            rc = rc_hdu.data

        try:
            flux_wcs = wcs.WCS(flux_hdu.header)
        except wcs.wcs.InconsistentAxisTypesError as exc:
            crpix1 = flux_hdu.header.get('CRPIX1')
            crval1 = flux_hdu.header.get('CRVAL1') 
            cdelt1 = flux_hdu.header.get('CDELT1') 

            if any([crpix1 is None, crval1 is None, cdelt1 is None]):
                raise exc

            # Fixing broken WCS
            flux_hdu.header['CTYPE1'] = 'WAVE'
            flux_hdu.header['Cunit1'] = 'Angstrom'
            
            flux_wcs = wcs.WCS(flux_hdu.header)


        if flux.shape != var.shape:
            raise ValueError(
                "Spectral data invalid or corruptede: Flux data shape "
                "do not match variance data one!"
            )

        # NOTE: Wavelenghts must be in Angstrom units
        pixel = np.arange(len(flux))
        if flux_wcs.has_spectral:
            lam = flux_wcs.pixel_to_world(pixel).Angstrom
        else:
            try:
                coeff0 = flux_hdu.header["COEFF0"]
                coeff1 = flux_hdu.header["COEFF1"]
            except KeyError:
                raise ValueError("Catton determine wavelength mapping")
            lam = 10**(coeff0 + coeff1*pixel)
        flux = flux.astype('float32')

        flux_not_nan_mask = ~np.isnan(flux)

        if mask is not None:
            flux_not_nan_mask &= ~(mask.astype(bool))

        flux_units = units.Unit(flux_hdu.header['BUNIT'])
        wave_units = units.Unit(flux_hdu.header['CUNIT1'])

        flux = flux.copy() * flux_units
        var = var.copy() * (flux_units**2)
        lam = lam.copy() * wave_units

        if rc is not None:
            rc = rc.copy()
        else:
            # If now wavelenght dispersion information is present, then
            # compute it using the wavelenght
            delta_lambda = np.ones_like(lam)
            delta_lambda[1:] = (lam[1:] - lam[:-1])
            delta_lambda = delta_lambda.to('Angstrom').value
            rc = resolution / delta_lambda
            rc[0] = rc[1]
        rc[rc < 1e-3] = 2.

        meta = {'header': hdulist[0].header}
        uncertainty = VarianceUncertainty(var)

        sp: Spectrum1D = Spectrum1D(
            flux=flux,
            spectral_axis=lam,
            wcs=flux_wcs,
            uncertainty=uncertainty,
            meta=meta
        )

        sp.mask = flux_not_nan_mask
        sp.rc = rc
        sp.obj_id = obj_id

    return sp

def read_sdss_extra(
    file: str
) -> Tuple[Union[np.ndarray, None], Union[np.ndarray, None]]:
    try:
        myt = Table.read(file, hdu='COADD')
    except KeyError:
        return None, None

    try:
        wd = myt['wdisp']
    except KeyError:
        wd = None

    try:
        sky = myt['sky']
    except KeyError:
        sky = None

    return sky, wd

def read(file: str) -> Spectrum1D:
    sp_formats = identify_spectrum_format(file)
    sp = Spectrum1D.read(file)
    sp.wd = None
    sp.sky = None

    # Try to read auxiliary data
    if 'SDSS-V spec' in sp_formats:
        sky, wd = read_sdss_extra(file)

        if sky is not None:
            sp.sky = sky * sp.flux.unit
        if wd is not None:
            sp.wd = ((10**wd)*units.Unit('Angstrom')).to(sp.spectral_axis.unit)

    return sp



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPEX - SPectra EXtractor.

Extract spectra from spectral data cubes.

Copyright (C) 2022  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
from __future__ import absolute_import, division, print_function

import os
import time
import shutil

import pytest
from pytestqt.qtbot import QtBot
from pytestqt.qt_compat import qt_api

from test import TEST_DATA_PATH
from redmost import gui

Z_FTOL = 0.01

TEST_FILE_95 = os.path.join(TEST_DATA_PATH, 'spec_95.fits')
PROJECT_1_FILE = os.path.join(TEST_DATA_PATH, 'test_project_1.json')
PROJECT_2_FILE = os.path.join(TEST_DATA_PATH, 'test_project_2.json')
ZCAT_FILE = os.path.join(TEST_DATA_PATH, 'test_zcat.fits')
SCREEN_OUT_DIR = os.path.join(TEST_DATA_PATH, 'screenshots')

if not os.path.isdir(SCREEN_OUT_DIR):
    os.makedirs(SCREEN_OUT_DIR)


def static_vars(**kwargs):
    def decorate(func):
        for k in kwargs:
            setattr(func, k, kwargs[k])
        return func
    return decorate

@static_vars(counter=0)
def save_screen(qtbot: QtBot, widget: qt_api.QtWidgets.QWidget):
    input_file = qtbot.screenshot(widget)
    file, ext = os.path.splitext(os.path.basename(input_file))
    dest_file_name = f"{file}_{save_screen.counter:04d}{ext}"
    shutil.copy(
        input_file,
        os.path.join(SCREEN_OUT_DIR, dest_file_name)
    )
    save_screen.counter += 1


def handle_message_box(button_type, qtbot: QtBot, main_app: gui.GuiApp):
    qtbot.wait_exposed(main_app.msgBox, timeout=1000)
    button = main_app.msgBox.button(button_type)
    shutil.copy(
        qtbot.screenshot(main_app.msgBox),
        SCREEN_OUT_DIR
    )
    assert button is not None
    qtbot.mouseClick(
        button,
        qt_api.QtCore.Qt.MouseButton.LeftButton,
        delay=1
    )

@pytest.fixture
def main_app(qtbot: QtBot) -> gui.GuiApp:
    print("Creating main window")
    main_app = gui.GuiApp()
    setattr(main_app.main_wnd, "closeEvent", lambda x: None)
    qtbot.addWidget(main_app.main_wnd)
    qtbot.addWidget(main_app.about_wnd)
    qtbot.addWidget(main_app.msgBox)
    return main_app


def test_new_project(qtbot: QtBot, main_app: gui.GuiApp):
    print("Creating a new project")
    main_app.main_wnd.action_new_project.trigger()
    save_screen(qtbot, main_app.main_wnd)


def test_load_project(qtbot: QtBot, main_app: gui.GuiApp):
    print("Loading a project")
    main_app.openProject(PROJECT_2_FILE)
    save_screen(qtbot, main_app.main_wnd)

def test_load_project_import_zcat(qtbot: QtBot, main_app: gui.GuiApp):
    print("Loading a project")
    main_app.openProject(PROJECT_1_FILE)
    main_app._doImportZcat(ZCAT_FILE)
    qtbot.wait_exposed(main_app.zcat_mapping_dialog)
    main_app.zcat_mapping_dialog.id_combo_box.setCurrentIndex(1)
    main_app.zcat_mapping_dialog.z_combo_box.setCurrentIndex(2)
    main_app.zcat_mapping_dialog.qf_combo_box.setCurrentIndex(4)
    save_screen(qtbot, main_app.zcat_mapping_dialog)
    main_app.zcat_mapping_dialog.accept()
    save_screen(qtbot, main_app.main_wnd)

def test_save_picture(qtbot: QtBot, main_app: gui.GuiApp):
    print("Open file")
    main_app.importSpectra([TEST_FILE_95, ])
    main_app._safe_set_spec_index(0)

    print("Set plot options")
    main_app.main_wnd.smoothing_check_box.setCheckState(
        qt_api.QtCore.Qt.CheckState.Checked
    )
    main_app.main_wnd.smoothing_dspinbox.setValue(8.5)
    main_app.main_wnd.show_lines_check_box.setCheckState(
        qt_api.QtCore.Qt.CheckState.Checked
    )

    print("Set redshift")
    main_app.main_wnd.z_dspinbox.setValue(0.3998)

    print("Saving pictures")
    save_screen(qtbot, main_app.main_wnd)
    main_app.exportAsPicture(
        os.path.join(SCREEN_OUT_DIR, f"spec_95.png")
    )



"""
Abstract apy to load different Qt backends

Note: Based on https://github.com/pytest-dev/pytest-qt
"""
from __future__ import annotations

import os
import sys
from typing import TYPE_CHECKING
from typing import Dict, Union, Optional, Any
from types import ModuleType
from collections import namedtuple

from redmost.utils import get_icon


if TYPE_CHECKING:
    # NOTE: These imports will be actually executed only by the type-checking
    #       programs, like mypy.
    try:
        from PyQt6 import QtCore, QtGui, QtWidgets, QtTest, QtCharts
        from PyQt6.QtCore import pyqtSignal as Signal
        from PyQt6.QtCore import pyqtSlot as Slot
        from PyQt6.QtCore import pyqtProperty as Property
    except (ModuleNotFoundError, ImportError):
        from PySide6 import QtCore, QtGui, QtWidgets, QtTest, QtCharts
        from PySide6.QtCore import Signal, Slot, Property


VersionTuple = namedtuple(
    typename="VersionTuple",
    field_names="qt_api, qt_api_version, runtime, compiled"
)

QT_APIS: Dict[str, str] = {
    x.lower(): x
    for x in ["PyQt6", "PySide6", "PyQt5", "PySide2"]
}


def _import(name):
    """Think call so we can mock it during testing"""
    return __import__(name)


def is_library_loaded(name):
    return name in sys.modules


def get_qt_api_from_env() -> Union[None, str]:
    api = os.environ.get("PYTEST_QT_API")
    if api is None:
        api = os.environ.get("QT_API")

    if api is not None:
        api = api.lower()
        if api not in QT_APIS.keys():  # pragma: no cover
            raise ValueError(
                f"Invalid value for $qt_api_name: {api}, "
                "expected one of {supported_apis}"
            )
    return api


def get_already_loaded_backend() -> Union[None, str]:
    for api, backend in QT_APIS.items():
        if is_library_loaded(backend):
            return api
    return None


class QtBackend:
    """
    Interface to the underlying Qt API currently configured.

    This object lazily loads all class references and other objects when
    the ``set_qt_api`` method gets called, providing a uniform way to
    access the Qt classes.
    """

    def __init__(self) -> None:
        self._import_errors = {}
        self.is_pyside: Union[None, bool] = None
        self.is_pyqt: Union[None, bool] = None
        self.qt_api_name: Union[None, str] = None

        self.QtCore: Optional[ModuleType] = None
        self.QtGui: Optional[ModuleType] = None
        self.QtTest: Optional[ModuleType] = None
        self.QtWidgets: Optional[ModuleType] = None
        self.QtCharts: Optional[ModuleType] = None

        self.qInfo: Optional[QtCore.qInfo] = None
        self.qDebug: Optional[QtCore.qDebug] = None
        self.qWarning: Optional[QtCore.qWarning] = None
        self.qCritical: Optional[QtCore.qCritical] = None
        self.qFatal: Optional[QtCore.qFatal] = None

        self.Signal: Optional[Signal] = None
        self.Slot: Optional[Slot] = None
        self.Property: Optional[Property] = None

        self.set_qt_api()

    def _guess_qt_api(self):  # pragma: no cover
        def _can_import(name):
            try:
                _import(name)
                return True
            except ModuleNotFoundError as e:
                self._import_errors[name] = str(e)
                return False

        # Note, not importing only the root namespace because
        # when uninstalling from conda, the namespace can still be there.
        for api, backend in QT_APIS.items():
            if _can_import(f"{backend}.QtCore"):
                return api
        return None

    def _import_module(self, module_name):
        _root_module = QT_APIS[self.qt_api_name]
        m = __import__(_root_module, globals(), locals(), [module_name], 0)
        return getattr(m, module_name)

    def set_qt_api(self, api=None):
        self.qt_api_name = (
            get_qt_api_from_env()
            or api
            or get_already_loaded_backend()
            or self._guess_qt_api()
        )

        self.is_pyside = self.qt_api_name in ["pyside2", "pyside6"]
        self.is_pyqt = self.qt_api_name in ["pyqt5", "pyqt6"]

        if not self.qt_api_name:  # pragma: no cover
            errors = "\n".join(
                f"  {module}: {reason}"
                for module, reason in sorted(self._import_errors.items())
            )
            raise ValueError(
                "Redmost requires either PySide2, PySide6, PyQt5 or PyQt6 "
                "installed.\n" + errors
            )

        self.QtCore = self._import_module("QtCore")
        self.QtGui = self._import_module("QtGui")
        self.QtTest = self._import_module("QtTest")
        self.QtWidgets = self._import_module("QtWidgets")

        if self.qt_api_name == "pyqt5":
            self.QtCharts = self._import_module("QtChart")
        else:
            self.QtCharts = self._import_module("QtCharts")

        self._check_qt_api_version()

        # qInfo is not exposed in PySide2/6 (#232)
        if hasattr(self.QtCore, "QMessageLogger"):
            self.qInfo = lambda msg: self.QtCore.QMessageLogger().info(msg)
        elif hasattr(self.QtCore, "qInfo"):
            self.qInfo = self.QtCore.qInfo
        else:
            self.qInfo = None

        self.qDebug = self.QtCore.qDebug
        self.qWarning = self.QtCore.qWarning
        self.qCritical = self.QtCore.qCritical
        self.qFatal = self.QtCore.qFatal

        if self.is_pyside:
            self.Signal = self.QtCore.Signal
            self.Slot = self.QtCore.Slot
            self.Property = self.QtCore.Property
        elif self.is_pyqt:
            self.Signal = self.QtCore.pyqtSignal
            self.Slot = self.QtCore.pyqtSlot
            self.Property = self.QtCore.pyqtProperty
        else:
            assert False, "Expected either is_pyqt or is_pyside"

    def _check_qt_api_version(self):
        pass
        # Just a placeholder for now
        #
        # if self.is_pyside:
        #    return
        # elif self.QtCore.PYQT_VERSION == 0x060000:  # 6.0.0
        #    return

    @staticmethod
    def exec(obj, *args, **kwargs):
        # exec was a keyword in Python 2, so PySide2 (and also PySide6 6.0)
        # name the corresponding method "exec_" instead.
        #
        # The old _exec() alias is removed in PyQt6 and also deprecated as of
        # PySide 6.1:
        # https://codereview.qt-project.org/c/pyside/pyside-setup/+/342095
        if hasattr(obj, "exec"):
            return obj.exec(*args, **kwargs)
        return obj.exec_(*args, **kwargs)

    def get_qicon(
        self,
        icon_name: str,
        icon_theme: Optional[str] = "feather"
    ) -> QtBackend.QtGui.QIcon:
        icon_file = get_icon(icon_name, icon_theme)
        if icon_file is None:
            return self.QtGui.QIcon()
        else:
            return self.QtGui.QIcon(icon_file)

    def get_versions(self):
        if self.qt_api_name == "pyside6":
            import PySide6

            version = PySide6.__version__

            return VersionTuple(
                "PySide6",
                version,
                self.QtCore.qVersion(),
                self.QtCore.__version__
            )
        elif self.qt_api_name == "pyside2":
            import PySide2

            version = PySide2.__version__

            return VersionTuple(
                "PySide2",
                version,
                self.QtCore.qVersion(),
                self.QtCore.__version__
            )
        elif self.qt_api_name == "pyqt6":
            return VersionTuple(
                "PyQt6",
                self.QtCore.PYQT_VERSION_STR,
                self.QtCore.qVersion(),
                self.QtCore.QT_VERSION_STR,
            )
        elif self.qt_api_name == "pyqt5":
            return VersionTuple(
                "PyQt5",
                self.QtCore.PYQT_VERSION_STR,
                self.QtCore.qVersion(),
                self.QtCore.QT_VERSION_STR,
            )

        raise ValueError(
            f"Internal error, unknown qt_api_name: {self.qt_api_name}"
        )

    def loadUiWidget(
        self,
        ui_filename: str,
        parent=None,
    ):
        """
        Load a UI file.

        :param ui_filename: Path of the UI file to load.
        :param parent: Parent widget. The default is None.
        :return ui: The widget loaded from the UI file.

        """

        ui_file_path: str = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), 'ui',
            ui_filename
        )

        if self.is_pyqt:
            uic = self._import_module("uic")
            ui = uic.loadUi(ui_file_path)
        elif self.is_pyside:
            uic = self._import_module("QtUiTools")
            loader = uic.QUiLoader()
            ui_file = self.QtCore.QFile(ui_file_path)
            ui_file.open(self.QtCore.QIODeviceBase.OpenModeFlag.ReadOnly)
            ui = loader.load(ui_file, parent)
            ui_file.close()
        else:
            raise ValueError("No valid GUI backend found!")

        return ui


qt_api = QtBackend()


def get_qapp():
    qapp = qt_api.QtWidgets.QApplication.instance()
    if qapp is None:
        # if it does not exist then a QApplication is created
        qapp = qt_api.QtWidgets.QApplication(sys.argv)
    return qapp
